"""
DifferencesROG.py
    A simple program to calculate the Radius of Gyration (ROG) of a
        bouncing beaded chain in a video, based on which pixels change
        from frame-to-frame. If the chain is not moving, this will not
        work.  If you have good contrast between the chain and the plate,
        SimpleROG.py may be a better choice since it does not depend on
        frame-to-frame differences.
        
    WARNING! This is new code, so it may have errors or bugs!
    
    To help you define the mask defining the plate where the chain is
        bouncing, the program displays the following images:
        - the first frame of the video with a pixel grid 
        - the current masked image of frame 0.
        - the last frame of the video 
        - the masked image of the last frame
        - the processed image of the difference between the last two frames.
    The program then plots the centre-of-mass and radius of gyration of the
        chain vs frame number.     

    Copyright (c) 2021 University of Toronto
    Original Version   :   28 March 2021
    Contact: David Bailey <dbailey@physics.utoronto.ca>
                (www.physics.utoronto.ca/~dbailey)
    License: Released under the MIT License; the full terms are
                found at www.opensource.org/licenses/mit-license.php
"""
 
## User input settings

# Input video file
#      Two sample videos are provided for testing: 'Knot_Dark_Plate.mp4', 'Chain_Al_Plate.mp4'
input_video          = 'Chain_Al_Plate.mp4'
# Size of pixel intensity difference that will be noticed
difference_threshold = 20
# Colour of grid that is useful for defining mask
grid_colour          = (255,0,0,255)  # RGB, intensity

# Mask centre and radius
# Be careful not to set radius too large. Since the plate is moving up and down,
#   the edge can appear and disappear if the mask is slightly too large,
#   which will introduce periodic artifacts in the ROG plot vs time.

mask_centre = ( 300 , 455 )
mask_radius =  250

## Libraries
import os
import imageio                    # conda install -c conda-forge imageio
from PIL import Image, ImageChops, ImageDraw, ImageFont # conda install -c anaconda pillow
import numpy
from scipy import ndimage

## Functions
def create_circular_mask(h, w, center=None, radius=None):
    # Creates a circular mask to select region of images over which 
    #    the Radius of Gyration will be calculated
    # Based on https://stackoverflow.com/questions/44865023
    if center is None: # use the middle of the image
        center = (int(w/2), int(h/2))
    if radius is None: # use the smallest distance between the center and image walls
        radius = min(center[0], center[1], w-center[0], h-center[1])

    Y, X = numpy.ogrid[:h, :w]
    dist_from_center = numpy.sqrt((X - center[0])**2 + (Y-center[1])**2)

    mask = dist_from_center <= radius
    return mask

def Display(frame) :
    try :
        # This displays image if being run from Jupyter notebook
        display(frame)
    except :
        # Otherwise saves the image
        frame.show()
    return

def draw_grid(frame, grid_steps=20, colour=(255,255,255,255)):
    # Draw a grid on a frame
    draw = ImageDraw.Draw(frame)
    y_start, y_end = 0, h
    step_size = int(w/ grid_steps)

    for x in range(0, w, step_size):
        line = ((x, y_start), (x, y_end))
        draw.line(line, fill=grid_colour)
        draw.text((x+3,3), str(x), fill=grid_colour)

    x_start, x_end = 0, w

    for y in range(0, h, step_size):
        line = ((x_start, y), (x_end, y))
        draw.line(line, fill=grid_colour)
        draw.text((3,y+3), str(y), fill=grid_colour)

    Display(frame)
    del draw


## Main Code Secion

# Create arrays to store centre position of chain and radius of gyration
chain_centre = []
chain_rog    = []

reader = imageio.get_reader(input_video)
for i, im in enumerate(reader):
    # Assumes light chain on dark plate.
    # Set lighter pixels to 0, darker pixels to 1
    # Create gray scale array by averaging rgb values in axis=2
    grey = im.mean(axis=2)
    # Convert to black and white image
    if i == 0 :
        h, w = im.shape[:2]
        print("frame height, width",h,w)
        frame = Image.fromarray(im)
        print("\nRaw Frame 0 of video with grid to help find centre of mask")
        draw_grid(frame)

        # Create circular mask for area of interest; note that center is (x,y)
        mask = create_circular_mask(h, w, center=mask_centre, radius=mask_radius)
        # Display masked frame 0
        masked_im = grey.copy()
        masked_im[~mask] = 128  # Set pixels outside mask to grey
        frame = Image.fromarray(masked_im.astype(numpy.uint8))
        print("\nCurrent masked grey-scale Frame 0")
        Display(frame)
        chain_centre.append((h/2,w/2))

    if 0 < i :
        masked_im = grey.copy()
        masked_im[~mask] = 128  # Set pixels outside mask to grey
        difference = numpy.abs(masked_im-previous_masked_im)
        difference = 255*(difference > difference_threshold)
        rog_data = difference.copy()       
        rog_data[~mask] = 0  # Set pixels outside mask to white
        chain_centre.append(ndimage.measurements.center_of_mass(rog_data))
        chain_rog.append(numpy.sqrt(((numpy.argwhere(rog_data>0) - chain_centre[i])**2).sum(1)).mean())        

    previous_masked_im = masked_im

print("\nLast frame of video")
Display(Image.fromarray(im))
print("\nMasked grey-scaled last frame of video")
Display(Image.fromarray(masked_im.astype(numpy.uint8)))
print("\nProcessed difference between last two frames of video")
Display(Image.fromarray(difference.astype(numpy.uint8)))       

# Plot data
print("\n\nInput Video:", input_video)

from matplotlib import pyplot

# Plot x,y centre of mass of chain
pyplot.figure(0)
fig1=pyplot.plot(chain_centre)
pyplot.suptitle("Chain Centre")
pyplot.xlabel("Frame Number")
pyplot.ylabel("x,y Centres (pixels)")
pyplot.legend(fig1, ('y', 'x'), loc='best')
pyplot.figure(1)

# Plot radius of gyration of chain
pyplot.suptitle("Chain Radius of Gyration")
pyplot.xlabel("Frame Number")
pyplot.ylabel("ROG (pixels)")
pyplot.plot(chain_rog)
pyplot.show()