"""
SimpleROG.py
    A simple program to calculate the Radius of Gyration (ROG) of a
        bouncing beaded chain in a video, assuming a light chain
        on dark plate.
    The code can be run standalone, or pasted into a Jupyter notebook cell.

    To help you define the mask defining the plate where the chain is
        bouncing, the program displays the following images:
        - the first frame of the video with a pixel grid 
        - the inverted first frame used to estimate the centre of the plate
            if the plate centre and radius not provided by user.
        - the masked processed black and white image used to calculate
            the centre and radius of gyration of the chain.
    
    Copyright (c) 2021 University of Toronto
    Original Version   :   28 March 2021
    Last Updated       :   14 April 2021 by Gabriel Bailey
    Contact: David Bailey <dbailey@physics.utoronto.ca>
                (www.physics.utoronto.ca/~dbailey)
    License: Released under the MIT License; the full terms are
                found at www.opensource.org/licenses/mit-license.php
"""
# Required imports
import os
import imageio        # conda install -c conda-forge imageio
from PIL import Image, ImageChops, ImageDraw, ImageFont # conda install -c anaconda pillow
import numpy as np
from scipy import ndimage, spatial
from datetime import datetime
from matplotlib import pyplot as plt
# If using lab computer anaconda may require admin password, so you can use pip instead:
# pip install imageio
# pip install imageio-ffmpeg
# pip install pillow
# pip install numpy
# pip isntall scipy
# pip isntall matplotlib

## User input settings
#===========================================================================================================================

# Input video file
#      Two sample videos are provided for testing: 'Knot_Dark_Plate.mp4', 'Chain_Al_Plate.mp4'
video_folder         = "../ROG Vids/"

# To process one video enter the name below:
input_video          = 'Video 25.wmv'

# Mask centre and radius
#   If set to None, they will be estimated
# Be careful not to set radius too large. Since the plate is moving up and down,
#   the edge can appear and disappear if the mask is slightly too large,
#   which will introduce periodic artifacts in the ROG plot vs time.
mask_centre     = 675,362
mask_radius     = 315

# The threshold of pixel intensity that will be assumed to be part of chain (can be values of 0(black) to 255(white))
colour_threshold    = 128

# =========== You may now run the program: the remaining settings are optional ===========



# This program processes light chains on a dark background. If your video is of the opposite, set this to true
invert = True

# To go over multiple videos with one run:
#   Set multi_vid to True, and the other two to a range that you want to process
#   Finally name_format must be a format string that represents the names of the videos
multi_vid       = False
start_vid_num   = 18
end_vid_num     = 39
name_format     = "Video {}.wmv" # will match 'Video 18.wmv', etc.

show_starter_pic    = True # Show a frame before analysis
show_bw_pic         = True # Show a b&w frame with a grid
show_final_pic      = True # Show the final frame used (should be all black besides the chain)
image_save          = False # Save the above images automatically (you might not be able to save them manually depending on OS)

# Note: showing plots will pause the program until they are closed.
# If you want to run this on multiple videos in the background, these must be false:
show_ROG_plot       = True # Show a plot of the ROG over time
show_COM_plot       = True # Show a plot of the COM position (helpful if you want to identify nonsymmetry in the plate or lighting)

write_ROG           = False # Writes the calculated ROGs in a space seperated file with their times
write_COM           = False # Writes the coordinates of the calculated COMs in a space seperated file (no time)
target_folder       = "." #The name of the folder to write to ("." if to the same folder it's being run in)
fID                 = "" # An identifier that will be placed at the end of the txt files' names e.g. "every 10 frames"

# Analysis parameters
#   step is how many frames you skip over to get each datapoint.
#       For every frame it should be 1, if you want to only analyze every 100th frame it should be 100.
#   start is the step (not frame!) you start analyzing on, end is the step you stop analyzing on
step    = 1
start   = 0
end     = 100 # set to a large enumber if you want to go to the end of the video: e.g. 10**1000000

# The length of the chain in cm
length_of_chain = 24.64

# Unit options:
in_seconds  = True          # Whether the output will be in seconds rather than frames
in_cm       = True          # Whether the output will be in cm rather than pixels
# Need these measurements for the conversion factors:
pan_width_cm    = 30.4      # The width of the pan in cm
pan_width_p     = 651.4     # The width of the pan in pixels
fps         = 15.0          # Framerate of the video


# Colour of grid that is useful for defining mask
grid_colour     = (255,0,0,255)  # RGB, intensity
#===========================================================================================================================

def create_circular_mask(h, w, center=None, radius=None):
    # Creates a circular mask to select region of images over which 
    #    the Radius of Gyration will be calculated
    # Based on https://stackoverflow.com/questions/44865023
    if center is None: # use the middle of the image
        center = (int(w/2), int(h/2))
    if radius is None: # use the smallest distance between the center and image walls
        radius = min(center[0], center[1], w-center[0], h-center[1])

    Y, X = np.ogrid[:h, :w]
    dist_from_center = np.sqrt((X - center[0])**2 + (Y-center[1])**2)

    mask = dist_from_center <= radius
    return mask

def Display(frame,name) :
    try :
        # This displays image if being run from Jupyter notebook
        display(frame)
    except :
        # Otherwise shows the image
        frame.show()
        if image_save:
            frame.save(target_folder+"/pic "+input_video[0:-4]+" "+name+".png")
    return

def draw_grid(img, grid_steps=20, colour=(255,255,255,255)):
    # Draw a grid on a frame
    cimg = Image.new("RGBA", img.size)
    cimg.paste(img)
    draw = ImageDraw.Draw(cimg)
    y_start, y_end = 0, h
    step_size = int(w/ grid_steps)

    for x in range(0, w, step_size):
        line = ((x, y_start), (x, y_end))
        draw.line(line, fill=grid_colour)
        draw.text((x+3,3), str(x), fill=grid_colour)

    x_start, x_end = 0, w

    for y in range(0, h, step_size):
        line = ((x_start, y), (x_end, y))
        draw.line(line, fill=grid_colour)
        draw.text((3,y+3), str(y), fill=grid_colour)

    Display(cimg, "grid")
    del draw
  
# Create arrays to store centre position of chain and radius of gyration
if not multi_vid:
    start_vid_num = 0
    end_vid_num = 0

for num in range(start_vid_num, end_vid_num+1):
    if multi_vid:
        input_video = name_format.format(num)

    chain_centre = []
    chain_rog    = []
    start_time = datetime.now()

    reader = imageio.get_reader(video_folder + input_video)

    i = start
    for frame, im in enumerate(reader):
        # Assumes light chain on dark plate.
        # Set lighter pixels to 0, darker pixels to 255
        # Create gray scale array by averaging rgb values in axis=2

        # Ignore frames outside of interval [start, end]
        if frame % step !=0 or frame < start*step:
            continue
        if frame > end*step:
            break
        
        grey = im.mean(axis=2)
        if show_starter_pic and (i == start):
            # Display unaltered image
            img1 = Image.fromarray(grey.astype(np.uint8))        
            Display(img1, "original")

        # Convert to black and white image
        if invert:
            bw = 255*(grey > colour_threshold)
        else:
            bw = 255*(grey < colour_threshold)
        if i == start:
            h, w = im.shape[:2]
            
            if show_bw_pic:
                # Display b&w image with grid
                img2 = Image.fromarray(bw)
                draw_grid(img2, grid_steps=50)
                 
            if mask_centre == None :
                print("mask_centre undefined")
                # Estimate rough centre of black plate by centre-of-mass of black pixels.
                com = ndimage.measurements.center_of_mass(bw)
                mask_centre = (com[1],com[0])
            if mask_radius == None :
                print("mask_radius undefined")
                # Estimate rough radius of plate from standard deviation of black pixels.
                mask_radius=1.5*max(np.std(np.nonzero(bw),axis=1))
                
            # Create circular mask for area of interest; note that center is (x,y)
            mask = create_circular_mask(h, w, center=mask_centre, radius=mask_radius)

        rog_data = 255-bw.copy()
        rog_data[~mask] = 0  # Set pixels outside mask to black

        if show_final_pic and i == start:
                # Display final processed image
                img3 = Image.fromarray(rog_data.astype(np.uint8))
                Display(img3, "final")

        chain_centre.append(list(ndimage.measurements.center_of_mass(rog_data)))
        
        chain_rog.append([i*step,np.sqrt(((np.argwhere(rog_data>0) - chain_centre[i-start])**2).sum(1)).mean()])

        i += 1

    # Plot data
    print("\n\nInput Video:", input_video)

    chain_rog = np.array(chain_rog)
    chain_centre = np.array(chain_centre)
    
    time, ROG = zip(*chain_rog)
    y, x = zip(*chain_centre)

    # Convert units if wanted
    if in_seconds:
        time = tuple(j / fps for j in time)

    if in_cm:
        ptocm = pan_width_cm/pan_width_p
        ROG = tuple(j * ptocm for j in ROG)
        x = tuple(j * ptocm for j in x)
        y = tuple(j * ptocm for j in y)

    # Print some basic info to the console
    print("Average ROG:     {0:2.3} +/- {1:2.3}".format(np.mean(ROG), np.std(ROG)))

    print("Average center:  ({0:2.3} +/- {1:2.3}, {2:2.3} +/- {3:2.3})".format(np.mean(x), np.std(x), np.mean(y), np.std(y)))

    # Write to the target files
    if write_ROG:
        np.savetxt(target_folder+"/"+input_video[0:-4]+" "+fID+" ROG.txt", chain_rog, delimiter=" ", header="# time, ROG")
    if write_COM:
        np.savetxt(target_folder+"/"+input_video[0:-4]+" "+fID+" COM.txt", chain_centre, delimiter=" ", header="# x position, y position")

    end_time = datetime.now()
    print("Time Taken: ", str(end_time - start_time))

    # Plot the data

    if show_COM_plot:
        plt.figure(0)
        plt.plot(x, y)
        plt.plot(np.mean(x),np.mean(y),"r+")

        plt.suptitle("Chain Centre")
        if in_cm:
            plt.xlabel("x position (cm)")
            plt.ylabel("y position (cm)")
        else:
            plt.xlabel("x position (pixels)")
            plt.ylabel("y position (pixels)")

    if show_ROG_plot:
        plt.figure(1)
        plt.plot(time,ROG)

        plt.suptitle("Chain Radius of Gyration")
        if in_seconds:
            plt.xlabel("Time (s)")
        else:
            plt.xlabel("Frame Number")
        if in_cm:
            plt.ylabel("ROG (cm)")
        else:
            plt.ylabel("ROG (pixels)")

    if show_COM_plot or show_COM_plot:
        plt.show()